<?php

namespace App\Controller;

use App\Entity\Category;
use App\Entity\Product;
use App\Repository\ProductRepository;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Routing\Annotation\Route;
use Symfony\Bundle\FrameworkBundle\Controller\Controller;
use Symfony\Component\HttpFoundation\Response;

class ProductController extends Controller
{

  /**
   * @Route(
   *  name = "product_index",
   *  path = "/product"
   * )
   *
   * @param \App\Repository\ProductRepository $repository
   *
   * @return \Symfony\Component\HttpFoundation\Response
   */
  public function index(ProductRepository $repository) {
    // Can also be fetched from the EntityManager, but simpler to inject it.
    $priceLimit = 21;
    /* Generated SQL */
$qb_SQL = <<<SQL
SELECT 
  p0_.description AS description_0, 
  p0_.id AS id_1, 
  p0_.name AS name_2, 
  p0_.price AS price_3, 
  p0_.category_id AS category_id_4 
FROM 
  product p0_ 
WHERE 
  p0_.price < ? 
ORDER BY 
  p0_.price ASC
SQL;

    $productsQB = $repository->findProductsLessThanQB($priceLimit);
    $productsDQL = $repository->findProductsLessThanDQL($priceLimit);
$plain_SQL = <<<SQL
SELECT 
  p.description, 
  p.id, 
  p.name, 
  p.price
FROM 
  product p
WHERE 
  p.price < :price
ORDER BY 
  p.price ASC;
SQL;

    $productsSQL = $repository->findProductsLessThanSQL($priceLimit);

    // These two build the exact same query object.
    $dql_qb = ($productsDQL === $productsQB);

    // But these only return the same raw results, missing the initialization.
    $dql_sql = ($productsDQL == $productsSQL);

    return $this->render('product/index.html.twig', [
      'products' => $productsQB,
      'dql_qb' => $dql_qb,
      'dql_sql' => $dql_sql,
    ]);
  }

  /**
   * @Route(
   *   name="product_add",
   *   path = "/product/add"
   * )
   */
  public function add() {
    // EntityManager can also be injected as EntityManagerInterface $em.
    $em = $this->getDoctrine()->getManager();

    $product = new Product();
    $product->setName('Keyboard')
      ->setPrice(19.99)
      ->setDescription('Ergonomic and stylish!');

    $cat = $em->getRepository(Category::class)->findOneBy([
      'name' => CategoryController::DEVICES,
    ]);
    $product->setCategory($cat);

    // tell Doctrine you want to (eventually) save the Product (no queries yet)
    $em->persist($product);

    // actually executes the queries (i.e. the INSERT query)
    $em->flush();
    $products = $cat->getProducts();

    return new Response('Saved new product with id ' . $product->getId(),
      Response::HTTP_CREATED);
  }

  /**
   * @Route(
   *   name = "product_show_manual",
   *   path = "/product/{id}/manual",
   * )
   *
   * @param int $id
   *
   * @return \Symfony\Component\HttpFoundation\Response
   */
  public function showManually(int $id) {
    $repository = $this->getDoctrine()
      ->getRepository(Product::class);

    // Find by PK (id).
    $product = $repository->find($id);

    // Find single entity by properties
    $product = $repository->findOneBy([
      'id' => $id,
    ]);

    $productsBy = $repository->findBy([
      'name' => 'Keyboard',
    ]);
    $this->addFlash('info', array_reduce($productsBy, function (string $accu, Product $product) {
      return $accu . '/' .  $product->getId() . ' ' . $product->getName() . ': ' . $product->getDescription();
    }, ''));

    if (!$product) {
      throw $this->createNotFoundException(
        "No product found for id $id"
      );
    }

    $cat_SQL = <<<SQL
SELECT 
  t0.id AS id_1, 
  t0.name AS name_2 
FROM 
  category t0 
WHERE 
  t0.id = ?
SQL;

    $category = $product->getCategory();

    /* How to remove a product using automatic orphans removal
    $category->removeProduct($product);
    $this->getDoctrine()->getManager()->flush();
    */

    return $this->render('product/product.html.twig', [
      'product' => $product,
      'category' => $category ? $category->getName() : '?',
    ]);
  }

  /**
   * @Route(
   *   name = "product_show_combined",
   *   path = "/product/{id}/combined",
   * )
   *
   * @param int $id
   *
   * @return \Symfony\Component\HttpFoundation\Response
   */
  public function showCombined(int $id, ProductRepository $repository) {
    $joined_SQL = <<<SQL
SELECT 
  p0_.description AS description_0, 
  p0_.id AS id_1, 
  p0_.name AS name_2, 
  p0_.price AS price_3, 
  c1_.id AS id_4, 
  c1_.name AS name_5, 
  p0_.category_id AS category_id_6 
FROM 
  product p0_ 
  INNER JOIN category c1_ ON p0_.category_id = c1_.id 
WHERE 
  p0_.id = ?
SQL;
    $product = $repository->findOneByIdJoinedToCategory($id);
    if (!$product) {
      throw $this->createNotFoundException(
        "No product found for id $id"
      );
    }

    // And at this point, no query, since category has been hydrated during the
    // joined query.
    $category = $product->getCategory();

    return $this->render('product/product.html.twig', [
      'product' => $product,
      'category' => $category->getName(),
    ]);
  }

  /**
   * @Route(
   *   name = "product_show",
   *   path = "/product/{id}",
   * )
   *
   * @param \App\Entity\Product $product
   *
   * @return \Symfony\Component\HttpFoundation\Response
   */
  public function show(Product $product) {
    return $this->render('product/product.html.twig', [
      'product' => $product,
      'category' => $product->getCategory()->getName(),
    ]);
  }

  /**
   * @Route(
   *   name = "product_raise",
   *   path = "/product/{id}/raise"
   * )
   *
   * @param \App\Entity\Product $product
   * @param \Doctrine\ORM\EntityManagerInterface $em
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   */
  public function update(Product $product, EntityManagerInterface $em) {
    $price = $product->getPrice();
    $price += 0.01;
    $product->setPrice($price);
    // $em->persist($product) is NOT needed but MAY be used as product is
    // already mapped.
    $em->flush();

    $id = $product->getId();
    $request = Request::create($this->generateUrl('product_show', ['id' => $id]));
    $kernel = $this->get('http_kernel');
    return $kernel->handle($request);
  }

  /**
   * @Route(
   *   name = "product_delete",
   *   path = "/product/{id}/delete"
   * )
   *
   * @param \App\Entity\Product $product
   * @param \Doctrine\ORM\EntityManagerInterface $em
   *
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   */
  public function delete(Product $product, EntityManagerInterface $em) {
    $em->remove($product);
    $em->flush();

    return $this->redirectToRoute('product_index');
  }
}
